#include "cgi-search.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "synonyms.h"


/*
*  list of synonyms
*
*  Input file structure
*  --------------------
*  Word <Single Tab or Space> Word <Line Feed>
*  E.G.:
*  color  <Single Tab or Space> colour <Line Feed>
*  colour <Single Tab or Space> color  <Line Feed>
*
*  Notes:
*	Each word needs to be in _BOTH_ columns!
*	First column needs to be alphabetically sorted using a C.UTF-8 locale,
*	    so that's unsigned ascending byte order.
*	Each word may occur only _ONCE_ per column!
*
*  List file structure
*  -------------------
*  Word <Null> Word <Null> ...
*  E.G.:
*  color <Null> colour <Null>
*
*  Index file structure
*  --------------------
*  Int Int ...
*  E.G.:
*  0  6
*  6  0
*
*  Each record is two four-byte signed integers.
*  E.G.:
*  'color' is at position 0 of the file
*  'colour' is at position 6 of the file
*  Position 6 is a synonym for position 0
*  Position 0 is a synonym for position 6
*/


/*  Words table  */
struct synwrd *wrdtab;
/*  Index table  */
struct synidx *idxtab;
/*  Record count; Nr of word - synonym pairs  */
int recnt;


/*  Compare orgs  */
static int comporg(const void *s1, const void *s2)
{
    struct synwrd *syn1, *syn2;

    syn1 = (struct synwrd *) s1;
    syn2 = (struct synwrd *) s2;

    return(strcmp(syn1->org, syn2->org));
}

/*  Find synonym in org column  */
int fndsyn(char *syn)
{
    struct synwrd key, *res;

    res = NULL;
    memset(&key, 0, sizeof(key));

    /*  Copy synonym to org (orgs are alphabetical, syns are not)  */
    strncpy(key.org, syn, CGS_WRD_MAX);
    /* void *bsearch(const void *key, const void *base, size_t nmemb, \
      size_t size, int (*comsyn)(const void *, const void *)); */
    res = bsearch(&key, wrdtab, recnt, sizeof(struct synwrd), comporg);
    if (res != NULL) {
	/*  Found; Return org number that goes with syn  */
	return(res->num);
    }

    return(-1);
}


int main(void)
{
    int i, j, len, lincnt, recptr, ret, size;
    char line[4096];
    FILE *idxfil, *lstfil, *synfil;

    lincnt = 0;		/*  Line count  */
    recnt  = 0;		/*  Record count  */
    recptr = 0;		/*  Pointer to record  */
    ret = 0;
    size = 0;
    idxfil = NULL;	/*  Index file  */
    idxtab = NULL;	/*  Index table  */
    lstfil = NULL;	/*  Synonym list file  */
    synfil = NULL;	/*  Synonym input file  */
    wrdtab = NULL;	/*  Word and their synonym table  */
    memset(line, 0, 4096);

    if ((synfil = fopen("synonyms.list", "r")) == NULL) {
	fprintf(stderr, "Cant open synonyms.list\n");
	exit(1);
    }

    /*  Count lines  */
    lincnt = 0;
    while (fgets(line, 4096, synfil)) {
	lincnt++;
    }
    /*  Add a margin  */
    lincnt++;

    /*  Rewind  */
    rewind(synfil);
    recnt = 0;
    /*  Allocate memory for word table  */
    size = lincnt * sizeof(struct synwrd);
    wrdtab = malloc((size_t) size);
    if (wrdtab == NULL ) {
	fprintf(stderr, "Can't allocate memory for wrdtab\n");
	fclose(synfil);
	exit(1);
    }
    memset(wrdtab, 0, size);
    while (fgets(line, 4096, synfil)) {
	i = 0;
	j = 0;
	len = 0;
	while (line[i] != 0 && i < 4096) {
	    /*
	    *  01234567
	    *  Foo Bar
	    *     ^^
	    *  len  j
	    */
	    if (line[i] == ' ' || line[i] == 9) {
		line[i] = 0;
		len = i;
		j = i + 1;
	    } else if (line[i] < ' ' && line[i] > -1) {
		line[i] = 0;
		break;
	    }
	    i++;
	}
	if (len > 1 && len < CGS_WRD_SIZ && i - j > 1 && i - j < CGS_WRD_SIZ) {
	    strncpy(wrdtab[recnt].org, line, CGS_WRD_MAX);
	    strncpy(wrdtab[recnt].syn, line + j, CGS_WRD_MAX);
	    wrdtab[recnt].num = recnt;	/*  Store word nr in struct  */
	    wrdtab[recnt].len = len;	/*  Length of word  */
	    recnt++;
	    if (recnt > lincnt) {
		fprintf(stderr, "Too many records: %d\n", recnt);
		free(wrdtab);
		fclose(synfil);
		exit(1);
	    }
	}
    }
    /*  Done with input file  */
    fclose(synfil);

    /*
    *  Add synonym index number to table
    *  0 1 color  colour
    *  1 0 colour color
    */
    i = 0;
    ret = -1;
    while (i < recnt) {
	/*  Find synonym in org column  */
	ret = fndsyn(wrdtab[i].syn);
	if (ret < 0) {
	    fprintf(stderr, "Synonym %s not found in org\n", wrdtab[i].syn);
	    free(wrdtab);
	    exit(1);
	}
	wrdtab[i].ptr = ret;
	i++;
    }

    /*  Allocate memory for idx table  */
    size = lincnt * sizeof(struct synidx);
    idxtab = malloc((size_t) size);
    if (idxtab == NULL) {
        fprintf(stderr, "Can't allocate memory for idxtab\n");
        free(wrdtab);
        exit(1);
    }

    /*  Synonyms list output file */
    if ((lstfil = fopen("synonyms-list", "w")) == NULL) {
        fprintf(stderr, "Cant open synonyms-list\n");
	free(idxtab);
	free(wrdtab);
        exit(1);
    }

    i = 0;
    recptr = 0;
    while (i < recnt) {
	/*  Write this string including terminating NULL  */
	fwrite(wrdtab[i].org, 1, wrdtab[i].len + 1, lstfil);
	/*  Offset in bytes to this string  */
	idxtab[i].org = recptr;
	/*  Next record  */
	recptr = recptr + wrdtab[i].len + 1;
	i++;
    }
    /*  Done with synonyms file  */
    fclose(lstfil);

    /*
    *  Add offset to synonym to table;
    *  Translate indexes to file offsets.
    */
    i = 0;
    while (i < recnt) {
	idxtab[i].syn = idxtab[wrdtab[i].ptr].org;
	i++;
    }

    if ((idxfil = fopen("synonyms.idx", "w")) == NULL) {
	fprintf(stderr, "Cant open synonyms.idx\n");
	free(idxtab);
	free(wrdtab);
	exit(1);
    }

    /*  Write index file  */
    i = 0;
    while (i < recnt)  {
	fwrite(&(idxtab[i]), sizeof(struct synidx), 1, idxfil);
	i++;
    }

    free(idxtab);
    free(wrdtab);
    fclose(idxfil);

    return(0);
}

